from __future__ import annotations
import json
from typing import Union, Dict, Any

class InvalidModelJSON(Exception):
    pass

def load_model_json(src: Union[str, dict]) -> Dict[str, Any]:
    """
    Load model JSON from a file path or a Python dict.
    Validates minimal required keys.
    """
    if isinstance(src, str):
        with open(src, "r", encoding="utf-8") as f:
            data = json.load(f)
    elif isinstance(src, dict):
        data = src
    else:
        raise InvalidModelJSON("Expected file path or dict.")

    if "layers" not in data or not isinstance(data["layers"], list):
        raise InvalidModelJSON("Model JSON must contain a 'layers' list.")
    # Backward-compatible defaults
    data.setdefault("id", "")
    data.setdefault("name", "OpenNetwork Model")
    return data



from __future__ import annotations
import argparse, json, sys
from .core import init as _init, setmodel as _setmodel, getmodel as _getmodel, ModelLoadError

def main():
    parser = argparse.ArgumentParser(prog="opennetwork", description="OpenNetwork Dev Suite CLI")
    sub = parser.add_subparsers(dest="cmd")

    p_init = sub.add_parser("init", help="Initialize and print quickstart banner")
    p_set = sub.add_parser("setmodel", help="Open the web playground")
    p_set.add_argument("--url", default=None, help="Override the playground URL")
    p_get = sub.add_parser("getmodel", help="Build a model from JSON and print a Keras summary")
    p_get.add_argument("path", help="Path to the model JSON")
    p_get.add_argument("--strict", action="store_true", help="Strict parsing (fail on unknowns)")
    p_get.add_argument("--no-auto-flatten", action="store_true", help="Disable auto-flatten before Dense")

    args = parser.parse_args()

    if args.cmd == "init":
        _init(verbose=True)
        return 0

    if args.cmd == "setmodel":
        _setmodel(url=args.url)
        return 0

    if args.cmd == "getmodel":
        try:
            model = _getmodel(args.path, strict=args.strict, auto_flatten=not args.no_auto_flatten)
            # Print summary to stdout
            model.summary()
            return 0
        except ModelLoadError as e:
            print(str(e), file=sys.stderr)
            return 2

    parser.print_help()
    return 0

if __name__ == "__main__":
    sys.exit(main())



from __future__ import annotations
import webbrowser
from typing import Optional

DEFAULT_URL = "https://opennetwork.dezors.com/playground"

def open_webview(url: Optional[str] = None, title: str = "OpenNetwork Playground", width: int = 1200, height: int = 800):
    """
    Try to open a lightweight native window (pywebview) if available,
    otherwise fall back to the default browser.
    """
    target = url or DEFAULT_URL
    try:
        import webview  # optional dependency
        webview.create_window(title, target, width=width, height=height)
        webview.start()
        return True
    except Exception:
        webbrowser.open(target, new=2)
        return False



from __future__ import annotations
from typing import Dict, Any, List, Set

class ONNError(Exception):
    pass

def toposort_layers(layers: List[dict]) -> List[dict]:
    """
    Topologically sort graph-like layers using their ids and outbound connections.
    Returns a *list of layer dicts* in dependency order.
    If there are multiple inputs/branches, keeps deterministic ordering by id.
    """
    id_to_node = {n["id"]: n for n in layers}
    indeg = {n["id"]: 0 for n in layers}
    for n in layers:
        for to_id in n.get("connections", []) or []:
            if to_id in indeg:
                indeg[to_id] += 1

    # queue starts with nodes with indegree 0
    queue = sorted([nid for nid, d in indeg.items() if d == 0])
    ordered_ids: List[str] = []
    while queue:
        nid = queue.pop(0)
        ordered_ids.append(nid)
        node = id_to_node[nid]
        for to_id in node.get("connections", []) or []:
            if to_id in indeg:
                indeg[to_id] -= 1
                if indeg[to_id] == 0:
                    queue.append(to_id)
        queue.sort()

    if len(ordered_ids) != len(layers):
        # There is a cycle or a dangling connection
        # We'll append any missing nodes in a stable order to avoid hard failure.
        remaining = [nid for nid in id_to_node.keys() if nid not in ordered_ids]
        ordered_ids.extend(sorted(remaining))

    return [id_to_node[i] for i in ordered_ids]

def ensure_list(x):
    if x is None:
        return []
    if isinstance(x, list):
        return x
    return [x]

def safe_get(d: dict, *keys, default=None):
    cur = d
    for k in keys:
        if isinstance(cur, dict) and k in cur:
            cur = cur[k]
        else:
            return default
    return cur



import json, os
import opennetwork as onn

def test_import():
    assert hasattr(onn, "getmodel")
    assert hasattr(onn, "setmodel")
    assert hasattr(onn, "init")

def test_dry_run_parsing():
    here = os.path.dirname(__file__)
    # use a small inline JSON
    graph = {
        "name": "Tiny",
        "layers": [
            {"id": "input-1", "type": "input", "params": {"input_shape": [28, 28, 1]}, "connections": ["conv-1"]},
            {"id": "conv-1", "type": "conv2d", "params": {"filters": 8, "kernel_size": [3,3]}, "connections": ["dense-1"]},
            {"id": "dense-1", "type": "dense", "params": {"units": 10}, "connections": []}
        ]
    }
    # Dry run path: return_extras True avoids TF dependency during tests
    model, extras = onn.getmodel(graph, return_extras=True)
    assert model is None
    assert "inserted" in extras



from __future__ import annotations
from typing import Union, Optional, Dict, Any
from .io import load_model_json
from .web import open_webview, DEFAULT_URL
from .parser import build_from_json
import logging

log = logging.getLogger("opennetwork")

class ModelLoadError(Exception):
    pass

def init(verbose: bool = True) -> bool:
    if verbose:
        banner = (
            "OpenNetwork • Dev Suite\n"
            " - setmodel() launches the Playground\n"
            " - getmodel(path_or_dict) builds a Keras model if TensorFlow is installed\n"
        )
        print(banner)
    return True

def setmodel(url: Optional[str] = None, **window_kwargs) -> bool:
    """
    Open a local webview (if pywebview is installed) or the default browser to the Playground.
    """
    return open_webview(url or DEFAULT_URL, **window_kwargs)

def getmodel(src: Union[str, dict], *, strict: bool = False, auto_flatten: bool = True, return_extras: bool = False):
    """
    Load the JSON and build a Keras model (if TensorFlow is available).
    Returns the tf.keras.Model. If TensorFlow is missing, raises ModelLoadError unless return_extras=True,
    in which case returns (None, {'graph_layers': [...], 'inserted': [...]}) for dry-run validation.
    """
    data = load_model_json(src)
    result = build_from_json(data, strict=strict, auto_flatten=auto_flatten)

    if result.model is None:
        msg = (
            "TensorFlow is not installed. Install it with:\n"
            "  pip install 'opennetwork[keras]'\n"
            "or add tensorflow manually: pip install tensorflow\n"
            "You can still validate the graph via return_extras=True."
        )
        if return_extras:
            return None, {"graph_layers": result.graph_layers, "inserted": result.inserted}
        raise ModelLoadError(msg)

    return result.model



from __future__ import annotations
from typing import Dict, Any, List, Tuple, Optional
from .utils import toposort_layers, safe_get
import logging

log = logging.getLogger("opennetwork")

LAYER_MAP = {
    "dense": "Dense",
    "conv2d": "Conv2D",
    "flatten": "Flatten",
    "dropout": "Dropout",
    "batchnorm": "BatchNormalization",
    "maxpool2d": "MaxPooling2D",
    "activation": "Activation",
    "softmax_output": "Activation",  # special-cased -> 'softmax'
    "input": "Input"
}

class BuildResult:
    def __init__(self, model=None, graph_layers: List[dict] = None, inserted: List[str] = None):
        self.model = model
        self.graph_layers = graph_layers or []
        self.inserted = inserted or []  # e.g., auto-added Flatten

def _maybe_import_tf():
    try:
        import tensorflow as tf  # noqa
        return tf
    except Exception as e:
        log.debug("TensorFlow not available: %s", e)
        return None

def _layer_from_spec(tf, spec: dict):
    tname = spec.get("type", "").lower()
    params = spec.get("params", {}) or {}

    if tname == "input":
        shape = tuple(params.get("input_shape") or [])
        return tf.keras.Input(shape=shape, name=spec.get("name") or spec.get("id"))

    if tname == "dense":
        return tf.keras.layers.Dense(
            units=int(params.get("units", 32)),
            activation=params.get("activation"),
            name=spec.get("name") or spec.get("id")
        )

    if tname == "conv2d":
        ks = params.get("kernel_size") or (3, 3)
        if isinstance(ks, list):
            ks = tuple(ks)
        strides = params.get("strides") or (1, 1)
        if isinstance(strides, list):
            strides = tuple(strides)
        padding = params.get("padding") or "valid"
        activation = params.get("activation")
        return tf.keras.layers.Conv2D(
            filters=int(params.get("filters", 32)),
            kernel_size=ks,
            strides=strides,
            padding=padding,
            activation=activation,
            name=spec.get("name") or spec.get("id")
        )

    if tname == "maxpool2d":
        ps = params.get("pool_size") or (2, 2)
        if isinstance(ps, list):
            ps = tuple(ps)
        strides = params.get("strides") or None
        if isinstance(strides, list):
            strides = tuple(strides)
        return tf.keras.layers.MaxPooling2D(pool_size=ps, strides=strides, name=spec.get("name") or spec.get("id"))

    if tname == "dropout":
        return tf.keras.layers.Dropout(rate=float(params.get("rate", params.get("dropout", 0.5))), name=spec.get("name") or spec.get("id"))

    if tname == "batchnorm":
        return tf.keras.layers.BatchNormalization(name=spec.get("name") or spec.get("id"))

    if tname == "flatten":
        return tf.keras.layers.Flatten(name=spec.get("name") or spec.get("id"))

    if tname == "activation":
        return tf.keras.layers.Activation(params.get("activation", "relu"), name=spec.get("name") or spec.get("id"))

    if tname == "softmax_output":
        return tf.keras.layers.Activation("softmax", name=spec.get("name") or spec.get("id"))

    raise ValueError(f"Unsupported layer type: {tname}")

def _graph_index(layers: List[dict]):
    by_id = {n["id"]: n for n in layers}
    parents = {n["id"]: [] for n in layers}
    children = {n["id"]: (n.get("connections") or [])[:] for n in layers}
    for n in layers:
        for c in n.get("connections") or []:
            if c in parents:
                parents[c].append(n["id"])
    return by_id, parents, children

def _needs_flatten(prev_shape: Tuple[int, ...]) -> bool:
    # If previous layer outputs rank > 2 and next is Dense, we need Flatten
    if prev_shape is None:
        return False
    return len(prev_shape) > 2

def build_from_json(data: Dict[str, Any], strict: bool = False, auto_flatten: bool = True) -> BuildResult:
    """
    Build a Keras Functional model from our JSON graph if tensorflow is available.
    If tensorflow is missing, returns a BuildResult with model=None and the sorted graph.
    """
    tf = _maybe_import_tf()
    layers = data.get("layers", [])
    ordered = toposort_layers(layers)
    inserted = []

    if tf is None:
        # Dry parse only
        return BuildResult(model=None, graph_layers=ordered, inserted=inserted)

    # Build functional API
    id_to_tensor = {}
    by_id, parents, children = _graph_index(ordered)

    # Create Input tensors first
    for n in ordered:
        if n.get("type") == "input":
            tensor = _layer_from_spec(tf, n)
            id_to_tensor[n["id"]] = tensor

    # Iterate in topological order to connect tensors
    for n in ordered:
        tname = n.get("type", "").lower()
        if tname == "input":
            continue

        inbound_ids = parents.get(n["id"], [])
        inbound_tensors = [id_to_tensor[i] for i in inbound_ids if i in id_to_tensor]

        # If no inbound tensors and not input, try to chain from last seen tensor
        x = None
        if inbound_tensors:
            if len(inbound_tensors) == 1:
                x = inbound_tensors[0]
            else:
                # If multiple inbound tensors, attempt to concatenate by default
                x = tf.keras.layers.Concatenate(name=f"{n['id']}_concat")(inbound_tensors)
        else:
            # If there is exactly one existing tensor, connect to it
            if len(id_to_tensor) == 1:
                x = list(id_to_tensor.values())[0]

        # Special handling: auto-flatten before Dense if rank>2
        if tname == "dense" and auto_flatten and x is not None:
            try:
                prev_shape = x.shape.as_list()[1:]  # exclude batch
                if _needs_flatten(tuple(prev_shape)):
                    x = tf.keras.layers.Flatten(name=f"{n['id']}_auto_flatten")(x)
                    inserted.append(f"Inserted Flatten before {n['id']}")
            except Exception:
                pass

        layer_obj = _layer_from_spec(tf, n)

        # If x is None (dangling), create dummy Input based on params if possible
        if x is None:
            if tname == "dense":
                # Fallback to a flat input of size units (best-effort)
                units = int(n.get("params", {}).get("units", 32))
                inp = tf.keras.Input(shape=(units,), name=f"{n['id']}_auto_input")
                x = inp
                inserted.append(f"Inserted Input({units}) for standalone Dense {n['id']}")
            else:
                # Generic fallback input (e.g., 28x28x1) if we can guess from any upstream input
                guessed = (28, 28, 1)
                inp = tf.keras.Input(shape=guessed, name=f"{n['id']}_auto_input")
                x = inp
                inserted.append(f"Inserted default Input{guessed} for {n['id']}")

        y = layer_obj(x) if callable(layer_obj) else layer_obj

        id_to_tensor[n["id"]] = y

    # Determine outputs: nodes with no children
    output_ids = [nid for nid, kids in children.items() if len(kids) == 0]
    outputs = [id_to_tensor[nid] for nid in output_ids if nid in id_to_tensor]

    # Inputs are tensors that are tf.keras.Input
    inputs = [t for t in id_to_tensor.values() if getattr(getattr(t, "_keras_history", None), "layer", None).__class__.__name__ == "InputLayer"]

    if not inputs:
        # Fallback: use first tensor as input if no explicit Input layer
        first = next(iter(id_to_tensor.values()))
        inputs = [first]

    if not outputs:
        # Fallback: last tensor is output
        outputs = [list(id_to_tensor.values())[-1]]

    model = tf.keras.Model(inputs=inputs, outputs=outputs, name=data.get("name", "OpenNetworkModel"))
    return BuildResult(model=model, graph_layers=ordered, inserted=inserted)
